@@ -2,11 +2,11 @@
 # this repository contains the full copyright notices and license terms.
 import stdnum.eu.vat as vat
 import stdnum.exceptions
-from sql import Null, Column
-from sql.functions import CharLength
+from sql import Null, Column, Literal
+from sql.functions import CharLength, Substring, Position
 
 from trytond.model import (ModelView, ModelSQL, MultiValueMixin, ValueMixin,
-    fields, Unique, sequence_ordered)
+    DeactivableMixin, fields, Unique, sequence_ordered)
 from trytond.wizard import Wizard, StateTransition, StateView, Button
 from trytond.pyson import Eval, Bool
 from trytond.transaction import Transaction
@@ -16,7 +16,8 @@ from trytond.tools.multivalue import migrate_property
 
 __all__ = ['Party', 'PartyLang', 'PartyCategory', 'PartyIdentifier',
     'CheckVIESResult', 'CheckVIES',
-    'PartyReplace', 'PartyReplaceAsk']
+    'PartyReplace', 'PartyReplaceAsk',
+    'PartyErase', 'PartyEraseAsk']
 
 VAT_COUNTRIES = [('', '')]
 STATES = {
@@ -25,7 +26,7 @@ STATES = {
 DEPENDS = ['active']
 
 
-class Party(ModelSQL, ModelView, MultiValueMixin):
+class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
     "Party"
     __name__ = 'party.party'
 
@@ -58,11 +59,6 @@ class Party(ModelSQL, ModelView, MultiValueMixin):
     categories = fields.Many2Many('party.party-party.category',
         'party', 'category', 'Categories', states=STATES, depends=DEPENDS,
         help="The categories the party belongs to.")
-    active = fields.Boolean('Active', select=True, states={
-            'readonly': Bool(Eval('replaced_by')),
-            },
-        depends=['replaced_by'],
-        help="Uncheck to exclude the party from future use.")
     replaced_by = fields.Many2One('party.party', "Replaced By", readonly=True,
         states={
             'invisible': ~Eval('replaced_by'),
@@ -84,14 +80,16 @@ class Party(ModelSQL, ModelView, MultiValueMixin):
              'The code of the party must be unique.')
         ]
         cls._order.insert(0, ('name', 'ASC'))
+        cls.active.states.update({
+                'readonly': Bool(Eval('replaced_by')),
+                })
+        cls.active.depends.append('replaced_by')
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-
         super(Party, cls).__register__(module_name)
 
-        table_h = TableHandler(cls, module_name)
+        table_h = cls.__table_handler__(module_name)
 
         # Migration from 3.8
         table_h.not_null_action('name', 'remove')
@@ -101,10 +99,6 @@ class Party(ModelSQL, ModelView, MultiValueMixin):
         table, _ = tables[None]
         return [CharLength(table.code), table.code]
 
-    @staticmethod
-    def default_active():
-        return True
-
     @staticmethod
     def default_categories():
         return Transaction().context.get('categories', [])
@@ -195,8 +189,9 @@ class Party(ModelSQL, ModelView, MultiValueMixin):
     def copy(cls, parties, default=None):
         if default is None:
             default = {}
-        default = default.copy()
-        default['code'] = None
+        else:
+            default = default.copy()
+        default.setdefault('code', None)
         return super(Party, cls).copy(parties, default=default)
 
     @classmethod
@@ -220,6 +215,7 @@ class Party(ModelSQL, ModelView, MultiValueMixin):
             ('code',) + tuple(clause[1:]),
             ('identifiers.code',) + tuple(clause[1:]),
             ('name',) + tuple(clause[1:]),
+            ('contact_mechanisms.rec_name',) + tuple(clause[1:]),
             ]
 
     def address_get(self, type=None):
@@ -227,20 +223,36 @@ class Party(ModelSQL, ModelView, MultiValueMixin):
         Try to find an address for the given type, if no type matches
         the first address is returned.
         """
-        Address = Pool().get("party.address")
-        addresses = Address.search(
-            [("party", "=", self.id), ("active", "=", True)],
-            order=[('sequence', 'ASC'), ('id', 'ASC')])
-        if not addresses:
-            return None
-        default_address = addresses[0]
-        if not type:
-            return default_address
-        for address in addresses:
-            if getattr(address, type):
-                return address
+        default_address = None
+        if self.addresses:
+            default_address = self.addresses[0]
+            if type:
+                for address in self.addresses:
+                    if getattr(address, type):
+                        return address
         return default_address
 
+    def contact_mechanism_get(self, types=None, usage=None):
+        """
+        Try to find a contact mechanism for the given types and usage, if no
+        usage matches the first mechanism of the given types is returned.
+        """
+        default_mechanism = None
+        if types:
+            if isinstance(types, str):
+                types = {types}
+            mechanisms = [m for m in self.contact_mechanisms
+                if m.type in types]
+        else:
+            mechanisms = self.contact_mechanisms
+        if mechanisms:
+            default_mechanism = mechanisms[0]
+            if usage:
+                for mechanism in mechanisms:
+                    if getattr(mechanism, usage):
+                        return mechanism
+        return default_mechanism
+
 
 class PartyLang(ModelSQL, ValueMixin):
     "Party Lang"
@@ -262,7 +274,7 @@ class PartyLang(ModelSQL, ValueMixin):
         super(PartyLang, cls).__register__(module_name)
 
         if not exist:
-            party_h = TableHandler(Party, module_name)
+            party_h = Party.__table_handler__(module_name)
             if party_h.column_exist('lang'):
                 query = table.insert(
                     [table.party, table.lang],
@@ -317,13 +329,12 @@ class PartyIdentifier(sequence_ordered(), ModelSQL, ModelView):
     def __register__(cls, module_name):
         pool = Pool()
         Party = pool.get('party.party')
-        TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
         party = Party.__table__()
 
         super(PartyIdentifier, cls).__register__(module_name)
 
-        party_h = TableHandler(Party, module_name)
+        party_h = Party.__table_handler__(module_name)
         if (party_h.column_exist('vat_number')
                 and party_h.column_exist('vat_country')):
             identifiers = []
@@ -418,7 +429,7 @@ class CheckVIES(Wizard):
                         parties_failed.append(party.id)
                     else:
                         parties_succeed.append(party.id)
-                except Exception, e:
+                except Exception as e:
                     if hasattr(e, 'faultstring') \
                             and hasattr(e.faultstring, 'find'):
                         if e.faultstring.find('INVALID_INPUT'):
@@ -447,7 +458,7 @@ class PartyReplace(Wizard):
     start_state = 'ask'
     ask = StateView('party.replace.ask', 'party.replace_ask_view_form', [
             Button("Cancel", 'end', 'tryton-cancel'),
-            Button("Replace", 'replace', 'tryton-find-replace', default=True),
+            Button("Replace", 'replace', 'tryton-launch', default=True),
             ])
     replace = StateTransition()
 
@@ -559,3 +570,139 @@ class PartyReplaceAsk(ModelView):
     def on_change_source(self):
         if self.source and self.source.replaced_by:
             self.destination = self.source.replaced_by
+
+
+class PartyErase(Wizard):
+    "Erase Party"
+    __name__ = 'party.erase'
+    start_state = 'ask'
+    ask = StateView('party.erase.ask', 'party.erase_ask_view_form', [
+            Button("Cancel", 'end', 'tryton-cancel'),
+            Button("Erase", 'erase', 'tryton-clear', default=True),
+            ])
+    erase = StateTransition()
+
+    @classmethod
+    def __setup__(cls):
+        super(PartyErase, cls).__setup__()
+        cls._error_messages.update({
+                'active_party': (
+                    'The party "%(party)s" can not be erased '
+                    'because he is still active.'),
+                })
+
+    def transition_erase(self):
+        pool = Pool()
+        Party = pool.get('party.party')
+        cursor = Transaction().connection.cursor()
+
+        def convert_from(table, tables):
+            right, condition = tables[None]
+            if table:
+                table = table.join(right, condition=condition)
+            else:
+                table = right
+            for k, sub_tables in tables.items():
+                if k is None:
+                    continue
+                table = convert_from(table, sub_tables)
+            return table
+
+        resources = self.get_resources()
+        parties = replacing = [self.ask.party]
+        with Transaction().set_context(active_test=False):
+            while replacing:
+                replacing = Party.search([
+                        ('replaced_by', 'in', list(map(int, replacing))),
+                        ])
+                parties += replacing
+        for party in parties:
+            self.check_erase(party)
+            to_erase = self.to_erase(party.id)
+            for Model, domain, resource, columns, values in to_erase:
+                assert issubclass(Model, ModelSQL)
+                assert len(columns) == len(values)
+                if 'active' in Model._fields:
+                    records = Model.search(domain)
+                    Model.write(records, {'active': False})
+
+                tables, where = Model.search_domain(domain, active_test=False)
+                from_ = convert_from(None, tables)
+                table, _ = tables[None]
+                query = from_.select(table.id, where=where)
+
+                if columns:
+                    model_tables = [Model.__table__()]
+                    if Model._history:
+                        model_tables.append(Model.__table_history__())
+                    for table in model_tables:
+                        sql_columns, sql_values = [], []
+                        for column, value in zip(columns, values):
+                            column = Column(table, column)
+                            sql_columns.append(column)
+                            sql_values.append(
+                                value(column) if callable(value) else value)
+                        cursor.execute(*table.update(
+                                sql_columns, sql_values,
+                                where=table.id.in_(query)))
+                if resource:
+                    for Resource in resources:
+                        model_tables = [Resource.__table__()]
+                        if Resource._history:
+                            model_tables.append(Resource.__table_history__())
+                        for table in model_tables:
+                            cursor.execute(*table.delete(
+                                    where=table.resource.like(
+                                        Model.__name__ + ',%')
+                                    & Model.id.sql_cast(
+                                        Substring(table.resource,
+                                            Position(',', table.resource) +
+                                            Literal(1))).in_(query)))
+        return 'end'
+
+    def check_erase(self, party):
+        if party.active:
+            self.raise_user_error('active_party', {
+                    'party': party.rec_name,
+                    })
+
+    def to_erase(self, party_id):
+        pool = Pool()
+        Party = pool.get('party.party')
+        Identifier = pool.get('party.identifier')
+        Address = pool.get('party.address')
+        ContactMechanism = pool.get('party.contact_mechanism')
+        return [
+            (Party, [('id', '=', party_id)], True,
+                ['name'],
+                [None]),
+            (Identifier, [('party', '=', party_id)], True,
+                ['type', 'code'],
+                [None, '****']),
+            (Address, [('party', '=', party_id)], True,
+                ['name', 'street', 'zip', 'city', 'country', 'subdivision'],
+                [None, None, None, None, None, None]),
+            (ContactMechanism, [('party', '=', party_id)], True,
+                ['value', 'name', 'comment'],
+                [None, None, None]),
+            ]
+
+    @classmethod
+    def get_resources(cls):
+        pool = Pool()
+        Attachment = pool.get('ir.attachment')
+        Note = pool.get('ir.note')
+        return [Attachment, Note]
+
+
+class PartyEraseAsk(ModelView):
+    "Erase Party"
+    __name__ = 'party.erase.ask'
+    party = fields.Many2One('party.party', "Party", required=True,
+        help="The party to be erased.")
+
+    @classmethod
+    def default_party(cls):
+        context = Transaction().context
+        if context.get('active_model') == 'party.party':
+            return context.get('active_id')
